(***********************************************************************)
(*                                                                     *)
(*                              Ref-indic                              *)
(*                                                                     *)
(*                        Refinement indicators                        *)
(*                                                                     *)
(*               EPI Pomdapi, INRIA Paris - Rocquencourt               *)
(*                                                                     *)
(*  Copyright (c) 2013-2014 INRIA.                                     *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved. This file is distributed only by permission.  *)
(*                                                                     *)
(*  Hend Ben Ameur <hbenameur@yahoo.ca>                                *)
(*  Francois Clement <Francois.Clement@inria.fr>                       *)
(*  Pierre Weis <Pierre.Weis@inria.fr>                                 *)
(*                                                                     *)
(***********************************************************************)

(* $Id: refinement.ml,v 1.47 2014/06/11 15:45:35 fclement Exp $ *)

(** {3 Refinement indicator algorithm}
 See algorithm in the outline of Ref-indic. *)

module Nat = Unatural;;
module Dgraph = Udynamic_graph;;
module Ddgraph = Udelayed_dgraph;;

module Set = Umatrix_nonempty_set;;
module Mapping = Umatrix_graph;;
module Structural_mapping = Umatrix_structural_mapping;;

module Config = Configuration;;

let r_qual = Upervasives.qualify "Refinement";;

let _r_failwith = Upervasives.failwith_qual r_qual;;
let r_print_debug fname = Config.print_debug (r_qual fname);;

open Refinement_types;;
open Refinement_types_print;;
open Scalar_spec;;
open Components_spec;;

module Make =
  functor (Scal : SCALAR) ->
  functor (Comps : COMPONENTS) ->
struct
  module D = Display.Make (Scal) (Comps);;
  module Bb = D.Bb;;
  module St = D.St;;
  module H = D.H;;
  
  let print_parameter = print_parameter Scal.print;;
  let print_gradient = print_gradient Scal.print;;
  let print_gradients = Comps.print print_gradient;;
  
  let ppf = Config.get_formatter ();;
  
  (* Iteration. *)
  
  (* Steps 2a and 2b. *)
  let zero_foi cc = { cc with first_order_indicator = Scal._0_; };;
  
  let zero_fois ({ Dgraph.image = ccl; _; } as zone) =
    { zone with Dgraph.image = Ulist.map zero_foi ccl; }
  ;;
  
  let same_cutting acc cc = acc.cutting == cc.cutting;;
  
  let add_foi2 acc cc =
    if not (same_cutting acc cc) then assert false else
    let afoi = acc.first_order_indicator
    and foi = cc.first_order_indicator in
    let afoi = Scal.( +.) afoi (Scal.square foi) in
    { acc with first_order_indicator = afoi; }
  ;;
  
  let same_zone_geometry z1 z2 =
    Parameter.same_geometry z1.Dgraph.value z2.Dgraph.value
  ;;
  
  let add_zone azone zone =
    if not (same_zone_geometry azone zone) then begin
      Format.fprintf ppf "@[<hv>azone =@ %a;@ zone =@ %a@.@]"
        print_parameter azone.Dgraph.value
        print_parameter zone.Dgraph.value;
      assert false end else
    let accl = Ulist.map2 add_foi2 azone.Dgraph.image zone.Dgraph.image in
    { azone with Dgraph.image = accl; }
  ;;
  
  (* state -> gradients -> forced_zonations *)
  let compute_candidate_cuttings (zs, _j) gs =
    let zs = Comps.map2 Ddgraph.do_it gs zs in
    let fzs = Comps.map Structural_mapping.to_array zs in
    match Config.get_dimension_management () with
    | Best_component_only -> fzs
    | Vector ->
        let fz0 = Comps.get fzs Unatural._0_ in
        let accu = Array.map zero_fois fz0 in
        let fz = Comps.fold_left (Uarray.map2 add_zone) accu fzs in
        Comps.make fz
  ;;
  
  (* Step 3. *)
  let dm_treatment f accu fzs =
    match Config.get_dimension_management () with
    | Best_component_only -> Comps.fold_left_i f accu fzs
    | Vector -> f accu Unatural._0_ (Comps.get fzs Unatural._0_)
  ;;
  
  let max_foi_cc accu { first_order_indicator = foi; _; } =
    max accu foi
  ;;
  
  let max_foi_zone accu { Dgraph.image = ccl; _; } =
    Ulist.fold_left max_foi_cc accu ccl
  ;;
  
  let max_foi_scalar accu _cn = Array.fold_left max_foi_zone accu;;
  
  let max_foi = dm_treatment max_foi_scalar Scal._0_;;
  
  let init_partition len zn (sst, ssf) fz i =
    let imax = pred (Unatural.to_int len) in
    let i' = if i = imax then zn else i in
    let fsmi' = Dgraph.get_value fz.(i') in
    let mi' = Mapping.get_constant fsmi' in
    let constant_sm ss = Mapping.constant ss.Set.nonempty_set mi' in
    if i = imax then constant_sm sst else
    if i = zn then constant_sm ssf else
    fsmi'
  ;;
  
  let split_scalar zn ct fz =
    let len = Unatural.succ (Uarray.length fz) in
    Uarray.init len (init_partition len zn ct fz)
  ;;
  
  let no_split_scalar fz = Array.map Dgraph.get_value fz;;
  
  let make_cutting cn zn ct =
    match Config.get_dimension_management () with
    | Best_component_only ->
        Comps.singleton cn (split_scalar zn ct) no_split_scalar
    | Vector ->
        Comps.init (fun _ -> split_scalar zn ct)
  ;;
  
  let split cn zn ct fzs =
    let cuttings = make_cutting cn zn ct in
    Comps.map_rails cuttings fzs
  ;;
  
  let select_cc
      fzs select cn zn accu
      { cutting = ct; first_order_indicator = foi; _; } =
    if select foi then (split cn zn ct fzs, foi) :: accu else accu
  ;;
  
  let select_zone fzs select cn accu zn fzone =
    Ulist.fold_left (select_cc fzs select cn zn) accu fzone.Dgraph.image
  ;;
  
  let select_scalar fzs select accu cn fz =
    Uarray.fold_left_i (select_zone fzs select cn) accu fz
  ;;
  
  (* forced_zonations -> (partitions * scalar indicator) list *)
  let select_candidate_cuttings fzs =
    let (ratio, nb) = Config.get_candidate_cuttings_selection_criterion () in
    let foi_max = max_foi fzs in
    let threshold = Scal.( *. ) (Scal.of_float ratio) foi_max in
    let select foi = foi >= threshold in
    let ppsl = dm_treatment (select_scalar fzs select) [] fzs in
    Ulist.map fst
      (Ulist.first (Unatural.to_int nb)
         (Ulist.sort
            (fun (_, foi1) (_, foi2) -> - (Pervasives.compare foi1 foi2))
            ppsl))
  ;;
  
  (* Step 4. *)
  (* state -> partitions list -> selected_cutting list *)
  let compute_selected_cuttings st ppsl = Ulist.map (Bb.optim st) ppsl;;
  
  (* Steps 5 and 6. *)
  (* state -> selected_cutting list -> state *)
  let select_optimal_cutting st ppsl scl =
    match (ppsl, scl) with
    | ([], []) -> st
    | ([], _ :: _) | (_ :: _, []) -> assert false
    | (_, sc0 :: scl') ->
        if Ulist.for_all2 St.equal_coarse_parameters ppsl scl then st else
        Ulist.fold_left
          (fun (_, aj as asc) (_, j as sc) -> if j < aj then sc else asc)
          sc0 scl'
  ;;
  
  (* history -> history *)
  let iteration sth_i =
    let i = H.head_state_number sth_i in
    let st_i = H.head_state sth_i in
    let gs_i = Bb.grad st_i in
    r_print_debug "iteration" "gs_i" print_gradients ppf gs_i;
    D.display_grad st_i gs_i;
    (* if small gradient then return sth_i else *)
    let fzs_i = compute_candidate_cuttings st_i gs_i in
    let ppsl_i = select_candidate_cuttings fzs_i in
    let scl_i = compute_selected_cuttings st_i ppsl_i in
    let ip1 = Unatural.succ i in
    let st_ip1 = select_optimal_cutting st_i ppsl_i scl_i in
    H.make_history_step ip1 st_ip1 sth_i
  ;;
  
  (* history -> bool *)
  let stop sth_i =
    let sc = Config.get_stopping_criterion () in
    let i = H.head_state_number sth_i in
    i >= sc.max_iteration (* || small relative cost decrease *)
  ;;
  
  (* history -> history *)
  let rec iterations sth_i =
    r_print_debug "iteration" "sth_i" H.print_head_h_state ppf sth_i;
    D.display_param sth_i;
    if stop sth_i then sth_i else
    let sth_ip1 = iteration sth_i in
    let stagnation = H.same_head_state sth_ip1 sth_i in
    if stagnation then sth_i else iterations sth_ip1
  ;;
  
  (* Algorithm. *)
  
  (* unit -> unit *)
  let algorithm () =
    let sth_1 = H.h_init in
    r_print_debug "algorithm" "sth_1" H.print_head_h_state ppf sth_1;
    D.init sth_1;
    let sth_n = iterations sth_1 in
    H.finalization sth_n;
  ;;
end
;;
